﻿'******************************
' アメダスデータダウンローダ、データコンバータに共通となる、初期設定周りのクラスです。
'
' [開発者]　Katsuhiro Morishita (Kumamoto-university Japan)
' [開発履歴]
'     2011/9/23   独立したモジュールとする
'     2011/9/24   何とか整備完了
'     2012/11/19  再整備
'                 設定リストにある観測所がリストになければ、エラー
'                 リストの二重登録もエラー
'     2012/11/22  エラーメッセージを更新
' *****************************
Imports System
Imports System.IO
Imports System.Text
Imports System.Text.RegularExpressions

Public Class AmedasSetting
    ' ****構造体**************************************************************************************
    ''' <summary>
    ''' 設定を保存する構造体
    ''' </summary>
    ''' <remarks></remarks>
    Private Structure setting
        ''' <summary>
        ''' 開始日
        ''' </summary>
        ''' <remarks></remarks>
        Dim startDate As Date
        ''' <summary>
        ''' 終了日
        ''' </summary>
        ''' <remarks></remarks>
        Dim stopDate As Date
        ''' <summary>
        ''' ダウンロードターゲット観測点番号
        ''' </summary>
        ''' <remarks></remarks>
        Dim targetList() As String
    End Structure
    ' ****メンバ変数**********************************************************************************
    ''' <summary>
    ''' アメダスのURLリストを格納したファイル名
    ''' </summary>
    ''' <remarks></remarks>
    Private ReadOnly AMEDAS_LIST As String = "AMEDAS list.csv"
    ''' <summary>
    ''' ユーザ設定ファイル名
    ''' </summary>
    ''' <remarks></remarks>
    Private ReadOnly SETTING_LIST As String = "AMEDAS.ini"
    ''' <summary>
    ''' 生のHTML保存用ディレクトリ名前
    ''' </summary>
    ''' <remarks></remarks>
    Private ReadOnly RAW_DATA_SAVE_DIR As String = "Raw HTML"
    ''' <summary>
    ''' HTMLを処理したファイルの保存ディレクトリ名前
    ''' </summary>
    ''' <remarks></remarks>
    Private ReadOnly PROCESSED_DATA_SAVE_DIR As String = "Processed HTML"
    ''' <summary>
    ''' ファイルの取得状況
    ''' <para>true: 正常に処理済みです。</para>
    ''' </summary>
    ''' <remarks></remarks>
    Private _isOpen As Boolean = False
    ''' <summary>
    ''' アメダスのコアデータ
    ''' </summary>
    ''' <remarks></remarks>
    Private amedasNode() As Amedas.Amedas
    ''' <summary>
    ''' ユーザ設定情報
    ''' </summary>
    ''' <remarks></remarks>
    Private info As setting
    ' ****プロパティ**********************************************************************************
    ''' <summary>
    ''' ファイルが正常に開いているかを返す
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property IsOpen As Boolean
        Get
            Return Me._isOpen
        End Get
    End Property
    ''' <summary>
    ''' 格納された設定ファイルに書かれた地点数を返す
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property Length As Integer
        Get
            If Me.IsOpen Then Return Me.info.targetList.Length Else Return 0
        End Get
    End Property
    ''' <summary>
    ''' 処理の開始日を返す
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property StartDate As Date
        Get
            If Me.IsOpen Then Return Me.info.startDate Else Return Nothing
        End Get
    End Property
    ''' <summary>
    ''' 処理の終了日を返す
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property StopDate As Date
        Get
            If Me.IsOpen Then Return Me.info.stopDate Else Return Nothing
        End Get
    End Property
    ' ****メソッド********************************************************************************************
    ''' <summary>
    ''' 指定されたインデックスに該当するアメダス情報を返す
    ''' </summary>
    ''' <param name="i">インデックス</param>
    ''' <returns>指定されたアメダス情報<para>該当のない場合は、初期化済みの情報構造体を返す。</para></returns>
    ''' <remarks></remarks>
    Public Function GetInformation(ByVal i As Integer) As Amedas.Amedas
        Dim ans As Amedas.Amedas

        ans = New Amedas.Amedas
        If Me.IsOpen And Me.Length > i And i >= 0 Then
            ans = Me.amedasNode(i)
        End If
        GetInformation = ans
    End Function
    ''' <summary>
    ''' HTMLファイル用に、アメダス観測点のID・地点名および時刻に基づく階層構造のフォルダを生成する
    ''' </summary>
    ''' <param name="day">日付</param>
    ''' <param name="i">インデックス</param>
    ''' <remarks>生成したフォルダまでのパスを返す<para>該当データが存在しない場合、""を返します。</para></remarks>
    Public Function CreateDirectory4rawHtmls(ByVal day As Date, ByVal i As Integer) As String
        Dim topdir As String = RAW_DATA_SAVE_DIR
        Dim target As String = ""
        Dim amedas As Amedas.Amedas

        If Me.IsOpen And Me.Length > i And i >= 0 Then                                                  ' 正常にファイルを開けたことと、指定された番号が正常かどうかを判定
            amedas = Me.amedasNode(i)                                                                   ' 指定要素のアメダスデータを代入（クラスなのでポインタのキャストに近いイメージ）
            target = topdir + "/" + amedas.ID.ToString + "_" + amedas.name + "/" + day.Year.ToString    ' ディレクトリ構造を生成
            If Not System.IO.Directory.Exists(target) Then System.IO.Directory.CreateDirectory(target) ' ディレクトリの存在を確認し、存在しない場合は生成する
        End If

        Return target
    End Function
    ''' <summary>
    ''' HTMLを加工済みのファイル用に、アメダス観測点のID・地点名および時刻に基づく階層構造のフォルダを生成する
    ''' </summary>
    ''' <param name="day">日付</param>
    ''' <param name="i">インデックス</param>
    ''' <remarks>生成したフォルダまでのパスを返す</remarks>
    Public Function CreateDirectory4processedHtmls(ByVal day As Date, ByVal i As Integer) As String
        Dim topdir As String = PROCESSED_DATA_SAVE_DIR
        Dim target As String = ""
        Dim amedas As Amedas.Amedas

        If Me.IsOpen And Me.Length > i And i >= 0 Then                                                  ' 正常にファイルを開けたことと、指定された番号が正常かどうかを判定
            amedas = Me.amedasNode(i)                                                                   ' 指定要素のアメダスデータを代入（クラスなのでポインタのキャストに近いイメージ）
            target = topdir + "/" + amedas.ID.ToString + "_" + amedas.name + "/" + day.Year.ToString    ' ディレクトリ構造を生成
            If Not System.IO.Directory.Exists(target) Then System.IO.Directory.CreateDirectory(target) ' ディレクトリの存在を確認し、存在しない場合は生成する
        End If

        Return target
    End Function
    ''' <summary>
    ''' アメダスの気象観測データ公開URL（10分間隔の方）を生成する
    ''' </summary>
    ''' <param name="t">日付</param>
    ''' <param name="i">要素番号</param>
    ''' <returns>URL</returns>
    ''' <remarks></remarks>
    Public Function GetURL_10min(ByVal t As Date, ByVal i As Integer)
        Dim ans As String = ""

        If Me.IsOpen And Me.Length > i And i >= 0 Then              ' 正常にファイルを開けたことと、指定された番号が正常かどうかを判定
            ans = Me.amedasNode(i).GetURL_10min(t)
        End If

        Return ans
    End Function
    ''' <summary>
    ''' HTMLファイル保存時のファイル名を返す
    ''' <para>何もこんなの作らなくてもと思うかもしれないが、保存したhtmlを解析するには必要なんです。</para>
    ''' </summary>
    ''' <param name="t">日付</param>
    ''' <param name="i">要素番号</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetFileName4Html(ByVal t As Date, ByVal i As Integer)
        Dim ans As String = ""
        Dim amedas As Amedas.Amedas

        If Me.IsOpen And Me.Length > i And i >= 0 Then                                                  ' 正常にファイルを開けたことと、指定された番号が正常かどうかを判定
            amedas = Me.amedasNode(i)                                                                   ' 指定要素のアメダスデータを代入
            ans = amedas.ID.ToString + "_" + amedas.name + "_" + t.Year.ToString + "_" + t.ToString("MM") + "_" + t.ToString("dd") + ".html"  ' ファイル名生成
        End If

        Return ans
    End Function
    ''' <summary>
    ''' HTMLを加工したファイルのファイル名を返す
    ''' </summary>
    ''' <param name="t">日付</param>
    ''' <param name="i">要素番号</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetFileName4Process(ByVal t As Date, ByVal i As Integer)
        Dim ans As String = ""
        Dim amedas As Amedas.Amedas

        If Me.IsOpen And Me.Length > i And i >= 0 Then                                                  ' 正常にファイルを開けたことと、指定された番号が正常かどうかを判定
            amedas = Me.amedasNode(i)                                                                   ' 指定要素のアメダスデータを代入
            ans = amedas.ID.ToString + "_" + amedas.name + "_" + t.Year.ToString + "_" + t.ToString("MM") + "_" + t.ToString("dd") + ".csv"  ' ファイル名生成
        End If

        Return ans
    End Function
    ''' <summary>
    ''' アメダスのリストファイルを確認し、存在していた場合は呼び出し元にリストをハッシュテーブルにより渡す
    ''' <para>ファイルのフォーマットは、都道府県・観測官署番号(Prec No.)とID（Block No.）と地点名とします。</para>
    ''' <para>例えば、「11,49125,盛岡」とする。なお、この番号と地点名はテキトウです。</para>
    ''' </summary>
    ''' <param name="hs">ハッシュテーブル</param>
    ''' <returns>ファイルが存在すればture</returns>
    ''' <remarks></remarks>
    Private Function GetAmedasList(ByRef hs As Hashtable) As Boolean
        Dim ans As Boolean = False
        Dim line As String

        If File.Exists(AMEDAS_LIST) Then                            ' ファイルの存在を確認
            ans = True
            'AMEDAS_URL_LISTをShift-JISコードとして開く
            Using sr As New System.IO.StreamReader(AMEDAS_LIST, Encoding.UTF8)
                Dim r As New Regex("(?<prec_no>\d{2})(,|\t)(?<block_no>\d+)(,|\t)(?<name>\w+)", RegexOptions.IgnoreCase) ' Regexオブジェクトを作成 
                While sr.EndOfStream = False
                    line = sr.ReadLine()                            ' ファイルから1行取得
                    Dim mc As Match = r.Match(line)
                    While mc.Success
                        Dim name As String = mc.Groups("name").Value
                        Dim id As String = mc.Groups("block_no").Value
                        Dim precNo As Integer = Integer.Parse(mc.Groups("prec_no").Value)
                        hs(id) = New Amedas.Amedas(id, name, precNo)
                        mc = mc.NextMatch()
                    End While
                End While
            End Using
        Else
            Console.WriteLine("アメダスリストエラー")
        End If
        GetAmedasList = ans                                         ' 返値を格納して終了
    End Function
    ''' <summary>
    ''' 収集開始日と終了日及び収集対象地点名を格納したファイルの存在を確認し、存在していた場合はデータを構造体に格納させる
    ''' <para>ファイルのフォーマットは、1行目に開始日、2行目に終了日、3行目以降にID（Block No.）と地点名を記す。</para>
    ''' <para>例えば、「49125,盛岡」とする。なお、この番号と地点名はテキトウです。</para>
    ''' </summary>
    ''' <param name="info">設定を保存する構造体</param>
    ''' <returns>問題なければture</returns>
    Private Function GetTargetList(ByRef info As setting) As Boolean
        Dim ans As Boolean = False
        Dim line As String
        Dim list As List(Of String) = New List(Of String)()

        info.startDate = New Date(2000, 1, 1)
        info.stopDate = New Date(2000, 1, 1)

        If File.Exists(SETTING_LIST) Then                               ' ファイルの存在を確認
            ans = True
            'AMEDAS_URL_LISTをShift-JISコードとして開く
            Using sr As New System.IO.StreamReader(SETTING_LIST, Encoding.UTF8)
                Dim cnt As Integer = 0
                Dim r As New Regex("(?<block_no>\d+)(,|\t)(?<name>\w+)", RegexOptions.IgnoreCase) ' Regexオブジェクトを作成 
                Dim field() As String

                While sr.EndOfStream = False
                    If cnt = 0 Then
                        line = sr.ReadLine                              ' 1行目取得
                        field = line.Split("=")(1).Split(",")
                        Dim year, month, day As Integer
                        year = Integer.Parse(field(0))
                        month = Integer.Parse(field(1))
                        day = Integer.Parse(field(2))
                        info.startDate = New Date(year, month, day)     ' 日付オブジェクトを生成
                    End If
                    If cnt = 1 Then
                        line = sr.ReadLine                              ' 2行目取得
                        field = line.Split("=")(1).Split(",")
                        Dim year, month, day As Integer
                        year = Integer.Parse(field(0))
                        month = Integer.Parse(field(1))
                        day = Integer.Parse(field(2))
                        info.stopDate = New Date(year, month, day)      ' 日付オブジェクトを生成
                    End If
                    If cnt >= 2 Then
                        line = sr.ReadLine                              ' 3行目以降を取得
                        Dim mc As Match = r.Match(line)
                        While mc.Success
                            Dim id As String = mc.Groups("block_no").Value
                            list.Add(id)
                            mc = mc.NextMatch()
                        End While
                    End If
                    cnt += 1
                End While
            End Using
        End If
        info.targetList = list.ToArray()                                ' ダウンロード対象となる観測点番号を配列で格納
        ' エラー対策
        If info.startDate > info.stopDate Then ans = False ' もしスタート日がストップ日よりも大きければ処理しない
        If info.targetList.Length = 0 Then ans = False ' 対象観測点数が0ならやはり処理しない
        If ans = False Then Console.WriteLine("設定ファイルエラー")
        GetTargetList = ans                                             ' 返値を格納して終了
    End Function
    ''' <summary>
    ''' コンストラクタ
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        Dim amedasList As Hashtable = New Hashtable                                 ' ハッシュテーブル（観測地点番号block noをキーにした、アメダス情報）
        Dim targetInfo As setting = New setting                                     ' 読み込んだ設定を格納する
        Dim i As Integer

        If Me.GetAmedasList(amedasList) And Me.GetTargetList(targetInfo) Then
            Me._isOpen = True                                                       ' 正常に設定ファイルを読み込めたことを確認する
            Me.info = targetInfo
            Me.amedasNode = New Amedas.Amedas(targetInfo.targetList.Length - 1) {}  ' 用意するアメダスデータ配列のサイズ決定とインスタンス生成（要素の宣言の仕方がCと違うので気持ち悪い…というか気持ち悪くなってしまった。）
            For i = 0 To Me.amedasNode.Length - 1
                Dim id As String = Me.info.targetList(i)
                Me.amedasNode(i) = amedasList(id)
            Next
        End If
    End Sub

End Class
